import matplotlib.pyplot
import numpy as np
import copy
import re


class McStasMetaData:
    """
    Class for holding metadata for McStas dataset, is to be read from
    mccode.sim file.

    Attributes
    ----------
    info : dict
        Contains read strings from mccode.sim in key, value

    dimension : Int or List of Int
        Int for 1d data set with lenght of data, Array for 2d with each
        length

    component_name : str
        Name of component in McStas file

    filename : str
        Name of data file to read

    limits : List
        Limits for monitor, length=2 for 1d data and length=4 for 2d
        data, for example spatial or time limits for monitor

    title : str
        Title of monitor when plotting, placed above plot

    xlabel : str
        Text for xlabel when plotting

    ylabel : str
        Text for ylabel when plotting

    Methods
    -------
    add_info(key,value)
        Adds an element to the info dictionary

    extract_info()
        Unpacks the information in info to class attributes

    set_title(string)
        Overwrites current title

    set_xlabel(string)
        Overwrites current xlabel

    set_ylabel(string)
        Overwrites current ylabel
    """

    def __init__(self):
        """Creating a new instance, no parameters"""
        self.info = {}

        self.component_name = None
        self.parameters = None
        self.filename = None
        self.dimension = None
        self.limits = []

        self.xlabel = None
        self.ylabel = None
        self.zlabel = None
        self.title = None

        self.total_I = None
        self.total_E = None
        self.total_N = None

    def add_info(self, key, value):
        """Adding information to info dict"""
        self.info[key] = value

    def extract_info(self):
        """Extracting information from info dict to class attributes"""

        # Extract dimension
        if "type" in self.info:
            type_data = self.info["type"]
            if "array_0d" in type_data:
                self.dimension = 0
            if "array_1d" in type_data:
                type_data = type_data.split("(")[1]
                type_data = type_data.split(")")[0]
                self.dimension = int(type_data)
            if "array_2d" in type_data:
                self.dimension = []
                type_string1 = type_data.split(",")[0]
                type_string1 = type_string1.split("(")[1]
                self.dimension.append(int(type_string1))

                type_string2 = type_data.split(",")[1]
                type_string2 = type_string2.split(")")[0]
                self.dimension.append(int(type_string2))
        else:
            raise NameError("No type in mccode data section!")

        # Extract component name
        if "component" in self.info:
            self.component_name = self.info["component"].rstrip()

        if "Parameters" in self.info:
            self.parameters = self.info["Parameters"]

        # Extract filename
        if "filename" in self.info:
            self.filename = self.info["filename"].rstrip()
        else:
            # Monitors without output files do exist
            print("The component named \"" + self.component_name
                  + "\" had no data file and will not be loaded.")
            self.filename = ""

        # Extract limits
        self.limits = []
        if "xylimits" in self.info:
            # find the four numbers xmin, xmax, ymin, ymax
            temp_str = self.info["xylimits"]
            limits_string = temp_str.split()
            for limit in limits_string:
                self.limits.append(float(limit))

        if "xlimits" in self.info:
            # find the two numbers, xmin, xmax
            temp_str = self.info["xlimits"]
            limits_string = temp_str.split()
            for limit in limits_string:
                self.limits.append(float(limit))

        # Extract plotting labels and title
        if "xlabel" in self.info:
            self.xlabel = self.info["xlabel"].rstrip()
        if "ylabel" in self.info:
            self.ylabel = self.info["ylabel"].rstrip()
        if "zlabel" in self.info:
            self.zlabel = self.info["zlabel"].rstrip()
        if "title" in self.info:
            self.title = self.info["title"].rstrip()

        if "values" in self.info:
            value_list = self.info["values"]
            value_list = value_list.strip().split(" ")
            self.total_I = float(value_list[0])
            self.total_E = float(value_list[1])
            self.total_N = float(value_list[2])

    def set_title(self, string):
        """Sets title for plotting"""
        self.title = string

    def set_xlabel(self, string):
        """Sets xlabel for plotting"""
        self.xlabel = string

    def set_ylabel(self, string):
        """Sets ylabel for plotting"""
        self.ylabel = string

    def set_zlabel(self, string):
        """Sets zlabel for plotting"""
        self.zlabel = string

    def __repr__(self):
        string = "metadata object\n"
        if self.component_name is not None:
            string += "component_name: " + self.component_name + "\n"

        if self.filename is not None:
            string += "filename: " + str(self.filename) + "\n"

        if self.dimension is not None:
            if type(self.dimension) == int and self.dimension == 0:
                string += "0D data"
                if self.xlabel is not None:
                    string += " " + self.xlabel + "\n"
                if self.ylabel is not None:
                    string += " " + self.ylabel + "\n"
                if self.zlabel is not None:
                    string += " " + self.zlabel + "\n"

            elif type(self.dimension) == int and self.dimension != 0:
                string += "1D data of length " + str(self.dimension) + "\n"
                if self.limits is not None:
                    string += "  [" + str(self.limits[0]) + ": "
                    string += str(self.limits[1]) + "]"
                if self.xlabel is not None:
                    string += " " + self.xlabel + "\n"
                if self.ylabel is not None:
                    string += " " + self.ylabel + "\n"
                if self.zlabel is not None:
                    string += " " + self.zlabel + "\n"

            elif len(self.dimension) == 2:
                string += "2D data of dimension (" + str(self.dimension[0])
                string += ", " + str(self.dimension[1]) + ")\n"
                if self.xlabel is not None:
                    if self.limits is not None:
                        string += "  [" + str(self.limits[0]) + ": "
                        string += str(self.limits[1]) + "]"
                    string += " " + self.xlabel + "\n"

                if self.ylabel is not None:
                    if self.limits is not None and len(self.limits) == 4:
                        string += "  [" + str(self.limits[2]) + ": "
                        string += str(self.limits[3]) + "]"
                    string += " " + self.ylabel + "\n"

                if self.zlabel is not None:
                    string += " " + self.zlabel + "\n"

        if self.parameters is not None and len(self.parameters)>0:
            string += "Instrument parameters: \n"
            for key in self.parameters:
                string += " " + str(key) + " = "
                string += str(self.parameters[key]) + "\n"

        return string


class McStasPlotOptions:
    """
    Class that holds plotting options related to McStas data set

    Attributes
    ----------
    log : bool, default False
        To plot on logarithmic or not, standard is linear

    orders_of_mag : float, default 300
        If plotting on log scale, restrict max range to orders_of_mag
        below maximum value

    colormap : string, default jet
        Chosen colormap for 2d data, should be available in matplotlib

    show_colorbar : bool, default True
        Selects if colorbar should be shown or not

    cut_max : float, default 1
        Factor multiplied onto maximum data value to set upper plot limit

    cut_min : float, default 0
        Removes given fraction of the plot range from the lower limit

    x_limit_multiplier : float, default 1
        Multiplies x axis limits with factor, useful for unit changes

    y_limit_multiplier : float, default 1
        Multiplies y axis limits with factor, useful for unit changes

    custom_ylim_bottom : bool, default False
        Indicates whether a manual lower limit for y axis has been set

    custom_ylim_top : bool, default False
        Indicates whether a manual upper limit for y axis has been set

    custom_xlim_left : bool, default False
        Indicates whether a manual lower limit for x axis has been set

    custom_xlim_right : bool, default False
        Indicates whether a manual upper limit for x axis has been set

    Methods
    -------
    set_options(keyword arguments)
        Can set the class attributes using keyword options

    """

    def __init__(self, *args, **kwargs):
        """Setting default values for plotting preferences"""
        self.log = False
        self.orders_of_mag = 300
        self.colormap = "jet"
        self.show_colorbar = True
        self.cut_max = 1
        self.cut_min = 0
        self.x_limit_multiplier = 1
        self.y_limit_multiplier = 1

        self.custom_ylim_bottom = False
        self.custom_ylim_top = False
        self.custom_xlim_left = False
        self.custom_xlim_right = False

        self.top_lim = None
        self.bottom_lim = None
        self.left_lim = None
        self.right_lim = None

    def set_options(self, **kwargs):
        """
        Set custom values for plotting preferences

        Keyword arguments
        -----------------

        log : bool, default False
            To plot on logarithmic or not, standard is linear

        orders_of_mag : float, default 300
            If plotting on log scale, restrict max range to orders_of_mag
            below maximum value

        colormap : string, default jet
            Chosen colormap for 2d data, should be available in matplotlib

        show_colorbar : bool, default True
            Selects if colorbar should be shown or not

        cut_max : float, default 1
            Factor multiplied onto maximum data value to set upper plot limit

        cut_min : float, default 0
            Removes given fraction of the plot range from the lower limit

        x_limit_multiplier : float, default 1
            Multiplies x axis limits with factor, useful for unit changes

        y_limit_multiplier : float, default 1
            Multiplies y axis limits with factor, useful for unit changes

        bottom_lim : float
            Set manual lower limit for y axis

        top_lim : float
            Set manual upper limit for y axis

        left_lim : float
            Set manual lower limit for x axis

        right_lim : float
            Set manual upper limit for x axis

        """
        if "log" in kwargs:
            self.log = bool(kwargs["log"])

        if "orders_of_mag" in kwargs:
            self.orders_of_mag = kwargs["orders_of_mag"]
            if not isinstance(self.orders_of_mag, (float, int)):
                raise ValueError("orders_of_mag must be a number, got: "
                                 + str(self.orders_of_mag))

        if "colormap" in kwargs:
            all_colormaps = matplotlib.pyplot.colormaps()
            self.colormap = kwargs["colormap"]
            if self.colormap not in all_colormaps:
                raise ValueError("Chosen colormap not available in "
                                 + "matplotlib, was: "
                                 + str(self.colormap))

        if "show_colorbar" in kwargs:
            self.show_colorbar = bool(kwargs["show_colorbar"])

        if "cut_max" in kwargs:
            self.cut_max = kwargs["cut_max"]
            if not isinstance(self.cut_max, (float, int)):
                raise ValueError("cut_max has to be a number, was given: "
                                 + str(self.cut_max))

        if "cut_min" in kwargs:
            self.cut_min = kwargs["cut_min"]
            if not isinstance(self.cut_min, (float, int)):
                raise ValueError("cut_min has to be a number, was given: "
                                 + str(self.cut_min))

        if "x_axis_multiplier" in kwargs:
            self.x_limit_multiplier = kwargs["x_axis_multiplier"]
            if not isinstance(self.x_limit_multiplier, (float, int)):
                raise ValueError("x_limit_multiplier has to be a number, was "
                                 + "given: " + str(self.x_limit_multiplier))

        if "y_axis_multiplier" in kwargs:
            self.y_limit_multiplier = kwargs["y_axis_multiplier"]
            if not isinstance(self.y_limit_multiplier, (float, int)):
                raise ValueError("y_limit_multiplier has to be a number, was "
                                 + "given: " + str(self.y_limit_multiplier))

        if "top_lim" in kwargs:
            self.top_lim = kwargs["top_lim"]
            self.custom_ylim_top = True
            if not isinstance(self.top_lim, (float, int)):
                raise ValueError("top_lim has to be a number, was "
                                 + "given: " + str(self.top_lim))

        if "bottom_lim" in kwargs:
            self.bottom_lim = kwargs["bottom_lim"]
            self.custom_ylim_bottom = True
            if not isinstance(self.bottom_lim, (float, int)):
                raise ValueError("bottom_lim has to be a number, was "
                                 + "given: " + str(self.bottom_lim))

        if "left_lim" in kwargs:
            self.left_lim = kwargs["left_lim"]
            self.custom_xlim_left = True
            if not isinstance(self.left_lim, (float, int)):
                raise ValueError("left_lim has to be a number, was "
                                 + "given: " + str(self.left_lim))

        if "right_lim" in kwargs:
            self.right_lim = kwargs["right_lim"]
            self.custom_xlim_right = True
            if not isinstance(self.right_lim, (float, int)):
                raise ValueError("right_lim has to be a number, was "
                                 + "given: " + str(self.right_lim))

    def __repr__(self):

        string = "plot_options"

        string += " log: " + str(self.log) + "\n"
        if self.log:
            string += " orders_of_mag: " + str(self.orders_of_mag) + "\n"

        string += " colormap: " + str(self.colormap) + "\n"
        string += " show_colorbar: " + str(self.show_colorbar) + "\n"
        string += " cut_min: " + str(self.cut_min) + "\n"
        string += " cut_max: " + str(self.cut_max) + "\n"
        string += " x_limit_multiplier: " + str(self.x_limit_multiplier) + "\n"
        string += " y_limit_multiplier: " + str(self.y_limit_multiplier) + "\n"

        if self.custom_xlim_left:
            string += "manual x lower limit: " + str(self.left_lim)

        if self.custom_xlim_right:
            string += "manual x upper limit: " + str(self.right_lim)

        if self.custom_ylim_bottom:
            string += "manual y lower limit: " + str(self.bottom_lim)

        if self.custom_ylim_bottom:
            string += "manual y upper limit: " + str(self.top_lim)

        return string


class McStasData:
    """
    Class for holding full McStas dataset with data, metadata and
    plotting preferences

    Attributes
    ----------
    metadata : McStasMetaData instance
        Holds the metadata for the dataset

    name : str
        Name of component, extracted from metadata

    plot_options : McStasPlotOptions instance
        Holds the plotting preferences for the dataset

    Methods
    -------
    set_xlabel : string
        sets xlabel of data for plotting

    set_ylabel : string
        sets ylabel of data for plotting

    set_title : string
        sets title of data for plotting

    set_options : keyword arguments
        sets plot options, keywords passed to McStasPlotOptions method
    """

    def __init__(self, metadata):
        """
        Initialize a new McStas dataset, 4 positional arguments, pass
        xaxis as kwarg if 1d data

        Parameters
        ----------
        metadata : McStasMetaData instance
            Holds the metadata for the dataset
        """

        # attach meta data
        self.metadata = metadata
        # get name from metadata
        self.name = self.metadata.component_name
        # initialize PlotOptions
        self.plot_options = McStasPlotOptions()

        self.data_type = None
        self.original_data_location = None

    # Methods xlabel, ylabel and title as they might not be found
    def set_xlabel(self, string):
        self.metadata.set_xlabel(string)

    def set_ylabel(self, string):
        self.metadata.set_ylabel(string)

    def set_zlabel(self, string):
        self.metadata.set_zlabel(string)

    def set_title(self, string):
        self.metadata.set_title(string)

    def set_plot_options(self, **kwargs):
        self.plot_options.set_options(**kwargs)

    def set_data_location(self, data_location):
        self.original_data_location = data_location

    def get_data_location(self):
        return self.original_data_location

    def __str__(self):
        """
        Returns string with quick summary of data
        """

        string = "McStasData: "
        string += self.name + " "
        if type(self.metadata.dimension) == int and self.metadata.dimension == 0:
            string += "type: 0D "
        elif type(self.metadata.dimension) == int and self.metadata.dimension != 0:
            string += "type: 1D "
        elif len(self.metadata.dimension) == 2:
            string += "type: 2D "
        else:
            string += "type: other "

        if self.metadata.total_I is not None:
            string += " I:" + str(self.metadata.total_I)
        if self.metadata.total_E is not None:
            string += " E:" + str(self.metadata.total_E)
        if self.metadata.total_N is not None:
            string += " N:" + str(self.metadata.total_N)

        return string

    def __repr__(self):
        return "\n" + self.__str__()


class McStasDataBinned(McStasData):
    """
    Class for holding full McStas dataset with data, metadata and
    plotting preferences

    Attributes
    ----------
    metadata : McStasMetaData instance
        Holds the metadata for the dataset

    name : str
        Name of component, extracted from metadata

    Intensity : numpy array
        Intensity data [neutrons/s] in 1d or 2d numpy array, dimension in
        metadata

    Error : numpy array
        Error data [neutrons/s] in 1d or 2d numpy array, same dimensions as
        Intensity

    Ncount : numpy array
        Number of rays in bin, 1d or 2d numpy array, same dimensions as
        Intensity

    plot_options : McStasPlotOptions instance
        Holds the plotting preferences for the dataset

    Methods
    -------
    set_xlabel : string
        sets xlabel of data for plotting

    set_ylabel : string
        sets ylabel of data for plotting

    set_zlabel : string
        sets ylabel of data for plotting

    set_title : string
        sets title of data for plotting

    set_options : keyword arguments
        sets plot options, keywords passed to McStasPlotOptions method
    """

    def __init__(self, metadata, intensity, error, ncount, **kwargs):
        """
        Initialize a new McStas dataset, 4 positional arguments, pass
        xaxis as kwarg if 1d data

        Parameters
        ----------
        metadata : McStasMetaData instance
            Holds the metadata for the dataset

        intensity : numpy array
            Intensity data [neutrons/s] in 1d or 2d numpy array, dimension in
            metadata

        error : numpy array
            Error data [neutrons/s] in 1d or 2d numpy array, same dimensions
            as Intensity

        ncount : numpy array
            Number of rays in bin, 1d or 2d numpy array, same dimensions as
            Intensity

        kwargs : keyword arguments
            xaxis is required for 1d data
        """

        super().__init__(metadata)

        # three basic arrays from positional arguments
        if not isinstance(intensity, np.ndarray):
            raise ValueError("intensity should be numpy array!")
        if not isinstance(error, np.ndarray):
            raise ValueError("error should be numpy array!")
        if not isinstance(ncount, np.ndarray):
            raise ValueError("ncount should be numpy array!")

        self.Intensity = intensity
        self.Error = error
        self.Ncount = ncount

        if type(self.metadata.dimension) == int and self.metadata.dimension == 0:
            self.data_type = "Binned 0D"
        elif type(self.metadata.dimension) == int and self.metadata.dimension != 0:
            self.data_type = "Binned 1D"
            if "xaxis" in kwargs:
                self.xaxis = kwargs["xaxis"]
            else:
                raise NameError(
                    "ERROR: Initialization of McStasData done with 1d "
                    + "data, but without xaxis for " + self.name + "!")
        elif len(self.metadata.dimension) == 2:
            self.data_type = "Binned 2D"
        else:
            self.data_type = "Binned"


class McStasDataEvent(McStasData):
    """
    Class for holding McStas event dataset with data, metadata and
    plotting preferences. Usually data the first one million events
    is plotted.

    Attributes
    ----------
    metadata : McStasMetaData instance
        Holds the metadata for the dataset

    name : str
        Name of component, extracted from metadata

    Events : numpy array
        Event data

    plot_options : McStasPlotOptions instance
        Holds the plotting preferences for the dataset

    Methods
    -------
    set_xlabel : string
        sets xlabel of data for plotting

    set_ylabel : string
        sets ylabel of data for plotting

    set_zlabel : string
        sets zlabel of data for plotting

    set_title : string
        sets title of data for plotting

    set_options : keyword arguments
        sets plot options, keywords passed to McStasPlotOptions method
    """

    def __init__(self, metadata, events, **kwargs):
        """
        Initialize a new McStas event dataset, 2 positional arguments

        Parameters
        ----------
        metadata : McStasMetaData instance
            Holds the metadata for the dataset

        events : numpy array
            event data
        """

        super().__init__(metadata)

        # three basic arrays from positional arguments
        if not isinstance(events, np.ndarray):
            raise ValueError("events should be numpy array!")

        self.Events = events
        self.data_type = "Events"

        self.variables = self.metadata.info["variables"].strip()
        self.variables = self.variables.split()

        # Calculate I, E and N
        if "p" in self.variables:
            p_array = self.get_data_column("p")
            total_I = p_array.sum()
            total_E = np.sqrt((p_array ** 2).sum())
            total_N = len(p_array)

            self.metadata.total_I = total_I
            self.metadata.total_E = total_E
            self.metadata.total_N = total_N

        else:

            self.metadata.total_I = None
            self.metadata.total_E = None
            self.metadata.total_N = None

        self.labels = {"t": "t [s]",
                       "x": "x [m]",
                       "y": "y [m]",
                       "z": "z [m]",
                       "vx": "vx [m/s]",
                       "vy": "vy [m/s]",
                       "vz": "vz [m/s]",
                       "l": "wavelength [AA]",
                       "e": "energy [meV]",
                       "speed": "speed [m/s]",
                       "dx": "divergence x [deg]",
                       "dy": "divergence y [deg]"}

    def find_variable_index(self, axis, flag_info=None):
        """
        Returns variable index for given axis name

        Parameters:

        axis : str
            Name of desired axis

        flag_info: list
            List of flag names used for user variables in event data
        """
        if flag_info is not None:
            # If flag info given, use it to find user var string
            for index, flag in enumerate(flag_info):
                if axis == flag:
                    axis = f"U{index + 1}"

        return self.variables.index(axis)

    def scale_weights(self, factor):
        """
        Scales all event weights with given factor

        Parameters:

        factor : float
            Factor with which all weights are scaled
        """
        self.Events[:, self.find_variable_index("p")] *= factor

    def get_label(self, axis, flag_info=None):
        """
        Returns data label corresponding to given axis name

        Parameters:

        axis : str
            Name of parameter

        flag_info : list
            list of names for user variables in event data set
        """
        axis = axis.lower()

        if flag_info is not None:
            # If flag info given, use it to find user var string
            for index, flag in enumerate(flag_info):
                if axis == flag:
                    return f"User{index+1}: {flag}"

        if axis in self.labels:
            return self.labels[axis]
        else:
            return ""

    def get_data_column(self, axis, flag_info=None):
        """
        Returns data column corresponding to given axis name

        Parameters:

        axis : str
            Name of parameter

        flag_info : list
            list of names for user variables in event data set
        """

        m_n_const = 1.674927e-27
        h_const = 6.626068e-34

        if axis.lower() == "speed":
            # Convert velocity to speed (must be before l and e)
            vx = self.Events[:, self.find_variable_index("vx")]
            vy = self.Events[:, self.find_variable_index("vy")]
            vz = self.Events[:, self.find_variable_index("vz")]
            return np.sqrt(vx ** 2 + vy ** 2 + vz ** 2)

        elif axis.lower() == "l":
            # Convert speed to lambda
            speed = self.get_data_column("speed")
            lambda_meter = h_const / (m_n_const*speed)
            return lambda_meter*1E10

        elif axis.lower() == "e":
            # Convert speed to energy
            speed = self.get_data_column("speed")
            energy_joule = 0.5 * m_n_const * speed ** 2
            return energy_joule/1.60217663E-19*1E3

        elif axis.lower() == "dx":
            # Convert velocity to divergence x
            vx = self.Events[:, self.find_variable_index("vx")]
            vz = self.Events[:, self.find_variable_index("vz")]
            return np.arctan(vx/vz) * 180 / np.pi

        elif axis.lower() == "dy":
            # Convert velocity to divergence y
            vy = self.Events[:, self.find_variable_index("vy")]
            vz = self.Events[:, self.find_variable_index("vz")]
            return np.arctan(vy/vz) * 180 / np.pi

        else:
            index = self.find_variable_index(axis, flag_info=flag_info)
            return self.Events[:, index]

    def make_1d(self, axis1, n_bins=50, flag_info=None):
        """
        Bin event data along to given axis to create binned dataset

        Parameters:

        axis1 : str
            Name of parameter for binned axis

        n_bins : integer
            Number of bins for histogramming

        flag_info : list
            list of names for user variables in event data set
        """
        data = self.get_data_column(axis1, flag_info)
        label = self.get_label(axis1, flag_info)

        weights = self.get_data_column("p", flag_info)
        intensity, edges = np.histogram(data, bins=n_bins, weights=weights)
        error_squared, edges = np.histogram(data, bins=n_bins, weights=weights**2)
        error = np.sqrt(error_squared)
        ncount, edges = np.histogram(data, bins=n_bins)

        centers = edges[0:-1] + 0.5*(edges[1] - edges[0])

        metadata = copy.deepcopy(self.metadata)
        metadata.dimension = len(centers)
        metadata.info["type"] = "array_1d"
        metadata.limits = [centers[0], centers[-1]]

        total_I = np.sum(intensity)
        total_E = np.sqrt(error_squared.sum())
        total_N = np.sum(ncount)
        metadata.info["values"] = "{:2.6E} {:2.6E} {:2.6E}".format(total_I, total_E, total_N)
        metadata.total_I = total_I
        metadata.total_E = total_E
        metadata.total_N = total_N

        binned = McStasDataBinned(metadata, intensity=intensity,
                                  error=error, ncount=ncount, xaxis=centers)

        binned.set_title("Binned data generated from events")
        binned.set_xlabel(label)
        binned.set_ylabel("Intensity per bin [n/s]")

        return binned

    def make_2d(self, axis1, axis2, n_bins=100, flag_info=None):
        """
        Bin event data along to given axes to create binned dataset

        Parameters:

        axis1 : str
            Name of parameter for first axis

        axis2 : str
            Name of parameter for second axis

        n_bins : integer or list
            Number of bins for histogramming, can be list with two elements

        flag_info : list
            list of names for user variables in event data set
        """

        data1 = self.get_data_column(axis1, flag_info)
        label1 = self.get_label(axis1, flag_info)
        data2 = self.get_data_column(axis2, flag_info)
        label2 = self.get_label(axis2, flag_info)

        if isinstance(n_bins, list):
            n_bins.reverse()

        weights = self.get_data_column("p", flag_info)
        intensity, edges2, edges1 = np.histogram2d(data2, data1, bins=n_bins, weights=weights)
        error_squared, edges2, edges1 = np.histogram2d(data2, data1, bins=n_bins, weights=weights**2)
        error = np.sqrt(error_squared)
        ncount, edges2, edges1 = np.histogram2d(data2, data1, bins=n_bins)

        centers1 = edges1[0:-1] + 0.5*(edges1[1] - edges1[0])
        centers2 = edges2[0:-1] + 0.5*(edges2[1] - edges2[0])

        metadata = copy.deepcopy(self.metadata)
        metadata.dimension = [len(centers1), len(centers2)]
        metadata.info["type"] = "array_2d"
        metadata.limits = [centers1[0], centers1[-1], centers2[0], centers2[-1]]

        total_I = intensity.sum()
        total_E = np.sqrt(error_squared.sum())
        total_N = ncount.sum()
        metadata.info["values"] = "{:2.6E} {:2.6E} {:2.6E}".format(total_I, total_E, total_N)
        metadata.total_I = total_I
        metadata.total_E = total_E
        metadata.total_N = total_N

        binned = McStasDataBinned(metadata, intensity=intensity,
                                  error=error, ncount=ncount)
        binned.set_title("Binned data generated from events")
        binned.set_xlabel(label1)
        binned.set_ylabel(label2)

        return binned

    def __str__(self):
        """
        Returns string with quick summary of data
        """

        string = "McStasDataEvent: "
        string += self.name + " with "
        string += str(len(self.Events)) + " events."
        if "variables" in self.metadata.info:
            string += " Variables: "
            string += self.metadata.info["variables"].strip()

        return string

    def __repr__(self):
        return "\n" + self.__str__()

def parse_coordinates(line, keyword):
    # Extract the coordinates from the line
    match = re.search(r'\(([^)]+)\)', line)
    if match:
        coords = match.group(1).split(',')
        coords = [float(coord.strip()) for coord in coords]
        return {f'{keyword}_x': coords[0], f'{keyword}_y': coords[1], f'{keyword}_z': coords[2]}
    return {}

class ComponentData:
    def __init__(self, file_path):
        self.file_path = file_path
        self.data = None

    def read(self):

        components = {}

        current_component = None

        with open(self.file_path, 'r') as file:
            for line in file:
                line = line.strip()
                if line.startswith('COMPONENT'):
                    match = re.match(r'COMPONENT (\S+) = (\S+)', line)
                    if match:
                        component_name = match.group(1)
                        component_type = match.group(2)
                        current_component = {'component': component_type}
                        components[component_name] = current_component
                        current_component["parameters"] = {}
                elif '=' in line:
                    if current_component is not None:
                        key, value = line.split('=',1)
                        try:
                            value = float(value)
                        except:
                            pass
                        current_component["parameters"][key] = value
                elif line.startswith('AT'):
                    if current_component is not None:
                        current_component.update(parse_coordinates(line, 'AT'))
                        if line.endswith('ABSOLUTE'):
                            current_component['AT_relative'] = "ABSOLUTE"
                        else:
                            current_component['AT_relative'] = line.split('RELATIVE')[1].strip()
                elif line.startswith('ROTATED'):
                    if current_component is not None:
                        current_component.update(parse_coordinates(line, 'ROTATED'))
                        if line.endswith('ABSOLUTE'):
                            current_component['ROTATED_relative'] = "ABSOLUTE"
                        else:
                            current_component['ROTATED_relative'] = line.split('RELATIVE')[1].strip()

        self.data = components

        return components




import os
from datetime import datetime
import time
import json

TIME_FORMAT = "%d/%m/%Y %H:%M:%S" # Time format used for database

class BeamDumpDatabase:
    def __init__(self, name, path):
        """
        Database over location and run properties of beam dumps

        name : str
            Name of instrument for which this database is connected

        path : str
            input_path for instrument, database is placed there
        """
        self.name = name
        self.path = path

        self.data = {}

        self.database_path = os.path.join(self.path, self.name + "_db")
        if os.path.isdir(self.database_path):
            self.load_database(self.database_path)
        else:
            self.create_new_database(self.database_path)

    def create_new_database(self, path):
        """
        Creates directory for database, may be expanded in future
        """
        os.mkdir(path)

    def load_database(self, path):
        """
        Loads an existing database from disk

        path : str
            Path for the database to be loaded
        """
        dump_points = os.listdir(path)

        # Each dump point is a folder in the database root
        for dump_point in dump_points:
            dump_point_path = os.path.join(path, dump_point)
            if not os.path.isdir(dump_point_path):
                continue

            if dump_point not in self.data:
                self.data[dump_point] = {}

            # Each run is a json file in the dump_point folder
            runs = os.listdir(dump_point_path)
            for run in runs:
                if run.endswith(".json"):
                    run_path = os.path.join(dump_point_path, run)
                    dump = BeamDump.dump_from_JSON(run_path)

                    run_name = dump.data["run_name"]
                    if run_name not in self.data[dump_point]:
                        self.data[dump_point][run_name] = {}

                    tag = dump.data["tag"]
                    self.data[dump_point][run_name][tag] = dump

    def create_folder_for_dump_point(self, dump_point):
        """
        Adds folder to database for dump_point if it hasn't been made yet

        Returns path to the folder whether it was created here or existed
        """
        dump_point_path = os.path.join(self.database_path, dump_point)
        if not os.path.isdir(dump_point_path):
            os.mkdir(dump_point_path)

        return dump_point_path

    def load_data(self, expected_filename, data_folder_path, parameters, run_name, dump_point, comment):
        """
        Include MCPL file into database with given metadata

        Attempts to load MCPL file from McStas output. If the file does not
        exists, the method will return without adding anything to the database.

        expected_filename : str
            Filename given as McStas parameter, can include explicit double quotes

        data_folder_path : str
            Path to the data folder that contains the MCPL file

        parameters: dict
            dict with parameter names and values for this run

        run_name : str
            Specified run name for this run

        dump_point : str
            Name of component where MCPL cut ended, can start from here later

        comment : str
            Comment on the run that can be included in metadata
        """
        if not os.path.isdir(data_folder_path):
            # If the folder doesn't exist, skip search
            return

        # Sanitize expected_filename
        expected_filename = expected_filename.strip('"')
        expected_filename = os.path.split(expected_filename)[-1]
        expected_filename = expected_filename.split(".gz")[0]
        expected_filename = expected_filename.split(".mcpl")[0]

        # The system might or might not be able to compress the data, search for both
        possible_endings = [".mcpl", ".mcpl.gz"]

        for ending in possible_endings:
            file_path = os.path.join(data_folder_path, expected_filename + ending)
            if not os.path.isfile(file_path):
                continue

            if dump_point not in self.data:
                self.data[dump_point] = {}

            if run_name not in self.data[dump_point]:
                self.data[dump_point][run_name] = {}

            n_tags = len(self.data[dump_point][run_name])
            # Find relative path
            db_path = self.path
            rel_path = os.path.relpath(file_path, start=db_path)

            # Create dump with given metadata
            dump = BeamDump(data_path=rel_path, parameters=parameters,
                            dump_point=dump_point, run_name=run_name,
                            comment=comment, tag=n_tags)
            # Write to disk
            dump_path = self.create_folder_for_dump_point(dump_point)
            dump.dump_to_JSON(dump_path)

            self.data[dump_point][run_name][n_tags] = dump
            return dump

    def get_dump(self, point, run_name=None, tag=None):
        """
        Getter for dumps with specified point, run_name and tag
        """
        if point not in self.data:
            raise KeyError("The dump point '" + point + "' wasn't in the database.")

        if run_name is None and tag is None:
            return self.newest_at_point(point)

        if run_name not in self.data[point]:
            raise KeyError("The run_name '" + run_name + "' wasn't in the database "
                           + "at the dump_point '" + point + "'.")

        if tag is None:
            return self.newest_at_point(point, run=run_name)

        tag = int(tag) # Ensure tag is an integer for database lookup

        if tag not in self.data[point][run_name]:
            raise KeyError("The tag '" + str(tag) + "' was not in the database for"
                           + "run_name '" + run_name
                           + "at the dump_point '" + point + "'.")

        dump = self.data[point][run_name][tag]
        if not dump.file_present(self.path):
            raise RuntimeError("The dump datafile was not found in the expected location.")

        return dump

    def newest_at_point(self, point, run=None):
        """
        Gets newest dump at a given point and optionally a run_name

        point : str
            String with component name matching the dump point

        run : str
            String matching the run_name desired
        """

        if point not in self.data:
            raise KeyError("The dump point '" + point + "' wasn't in the database.")

        if run is None:
            # Collect all runs at the point for sort_by_time
            runs = self.data[point]
        else:
            # If a run name is specified, check it exists
            if run not in self.data[point]:
                raise KeyError("The run_name '" + run + "' wasn't in the database "
                               + "at the dump_point '" + point + "'.")

            # Collect all the runs in a dict for sort_by_time
            runs = {run: self.data[point][run]}

        return self.sort_by_time(runs, return_latest=True)

    def sort_by_time(self, runs, return_latest=False):
        """
        Sorts a given dict of runs and returns a list or the latest if return_latest is true

        The input data is given as a dictionary of runs matching the database structure
        with all the individual tags underneath the runs. It is usually used with all runs
        being from the same point, but not necessary for the method to work.

        runs : dict
            Dictionary with runs containing dictionary of tags pointing to dumps

        return_latest : bool
            If True only latest dump is returned, otherwise sorted list
        """
        time_to_run = {}
        for run in runs:
            for tag in runs[run]:
                dump = runs[run][tag]
                time_loaded = dump.data["time_loaded"]
                time_to_run[time_loaded] = dump

        sorted_times = sorted((time.strptime(d, TIME_FORMAT) for d in time_to_run.keys()), reverse=True)

        return_list = []
        for sorted_time in sorted_times:
            return_list.append(time_to_run[time.strftime(TIME_FORMAT, sorted_time)])

        if return_latest:
            for dump in return_list:
                if dump.file_present(self.path):
                    return dump
                else:
                    point = dump.data["dump_point"]
                    run_name = dump.data["run_name"]
                    tag = dump.data["tag"]
                    data_path = dump.data["data_path"]

                    print(f"Latest file had run_name '{run_name}' and tag '{tag}', but the file was "
                          f"not found in the expected location, and thus skipped.")
                    print("Expected path: ", data_path)

            raise RuntimeError("No beam dumps available.")

        return return_list

    def show_in_order(self, component_names):
        """
        Method to print content of database in order of component_names

        component_names : list
            List of strings for component names in the instrument
        """

        if len(self.data) == 0:
            print("No data in dump database yet. Use run_to method to create dump.")
            return

        print("Run point:")
        print(" ", "run_name".ljust(12), ":", "tag", ":", "time".ljust(20), ":", "comment")
        print("-"*60, "--- -- -")

        for name in component_names:
            if name in self.data:
                print(name + ":")
                dumps = self.sort_by_time(self.data[name])
                for dump in dumps:
                    if len(dump.data["parameters"]) < 2:
                        par_string = ": " + str(dump.data["parameters"])
                    else:
                        par_string = ""

                    print(" ", dump.data["run_name"].ljust(12), ":",
                          str(dump.data["tag"]).ljust(3), ":",
                          dump.data["time_loaded"].ljust(20), ":",
                          dump.data["comment"], par_string)

    def __repr__(self):
        """
        Basic repr of the data included in the database.
        """
        string = ""
        for point in self.data:
            string += point + ":\n"
            for run in self.data[point]:
                for tag in self.data[point][run]:
                    string += "  " + self.data[point][run][tag].data["run_name"] + "\n:"
                    string += "    " + str(self.data[point][run][tag]) + ":\n"

        return string


class BeamDump:
    def __init__(self, data_path, parameters, dump_point, run_name,
                 time_loaded=None, comment="", tag=0, record_name=""):
        """
        Class describing a beamdump made somewhere in an instrument

        Inputs for class matches what can be read from json file describing
        dump to facilitate easy recreation from file. Strips complex parameters
        into simple dict with par name and value for easy storage.

        data_path : str
            Path of MCPL datafile

        parameters : dict
            Parameters used to run simulation

        dump_point : str
            Name of component where beam was dumped

        run_name : str
            Specified run name for this dump

        time_loaded : str
            Optional, usually generated at first load, timestamp

        comment : str
            Comment for this beamdump

        tag : integer
            Tag for this beamdump to differentiate similar dumps

        record_name : str
            Unique filename combining run_name and tag
        """
        simple_parameters = {}
        # Convert complex parameter objects to simple key - value pairs
        for parameter in parameters:
            if hasattr(parameters[parameter], "value"):
                # The full parameter objects store their value in value attribute
                simple_parameters[parameter] = parameters[parameter].value
            else:
                simple_parameters[parameter] = parameters[parameter]

        # The fields in data must correspond to the input of the class
        self.data = {"data_path": data_path,
                     "dump_point": dump_point,
                     "parameters": simple_parameters,
                     "run_name": run_name,
                     "comment": comment,
                     "tag": tag,
                     "record_name": record_name}

        if time_loaded is None:
            self.data["time_loaded"] = datetime.now().strftime(TIME_FORMAT)
        else:
            self.data["time_loaded"] = time_loaded

    @classmethod
    def dump_from_JSON(cls, filepath):
        """
        Load dump from json file

        filepath : str
            Path to json file for load
        """
        with open(filepath, "r") as f:
            data = json.loads(f.read())

        return cls(**data) # Since the data fields correspond to class input

    def dump_to_JSON(self, folder_path):
        """
        Writes representation of the object to file

        folder_path : str
            Folder in which the dump will be placed
        """

        self.data["record_name"] = self.data["run_name"] + "_" + str(self.data["tag"])

        filepath = os.path.join(folder_path, self.data["record_name"] + ".json")
        if os.path.isfile(filepath):
            raise RuntimeError("Run with this destination and run_name: '"
                               + self.data["record_name"]
                               + "' already exists in BeamDumpDatabase.")

        with open(filepath, "w") as outfile:
            json.dump(self.data, outfile)

    def file_present(self, origin_path):
        """
        Checks whether the file is present and returns bool
        """
        return os.path.isfile(os.path.join(origin_path,self.data["data_path"]))

    def print_all(self):
        """
        Print all entries in data for dump
        """
        for key in self.data:
            print(key, ":",  self.data[key])

    def __repr__(self):
        """
        Simple representation of data in object
        """
        return str(self.data)




import os
import os.path
import io
import unittest
import unittest.mock
import datetime

from mcstasscript.interface.instr import McStas_instr
from mcstasscript.interface.instr import McXtrace_instr

from mcstasscript.tests.helpers_for_tests import WorkInTestDir

run_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.')

def setup_instr_no_path():
    """
    Sets up a neutron instrument without a package_path
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McStas_instr("test_instrument")

    os.chdir(current_work_dir)

    return instrument


def setup_x_ray_instr_no_path():
    """
    Sets up a X-ray instrument without a package_path
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McXtrace_instr("test_instrument")

    os.chdir(current_work_dir)

    return instrument


def setup_instr_root_path():
    """
    Sets up a neutron instrument with root package_path
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McStas_instr("test_instrument", package_path="/")

    os.chdir(current_work_dir)

    return instrument


def setup_x_ray_instr_root_path():
    """
    Sets up a X-ray instrument with root package_path
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McXtrace_instr("test_instrument", package_path="/")

    os.chdir(current_work_dir)

    return instrument


def setup_instr_with_path():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder.
    """

    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McStas_instr("test_instrument", package_path=dummy_path)

    os.chdir(current_work_dir)  # Return to previous workdir

    return instrument


def setup_x_ray_instr_with_path():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder.
    """

    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McXtrace_instr("test_instrument", package_path=dummy_path)

    os.chdir(current_work_dir)  # Return to previous workdir

    return instrument


def setup_instr_with_input_path():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder. In addition the input_path
    is set to a folder in the test directory using an absolute path.
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")
    input_path = os.path.join(THIS_DIR, "test_input_folder")

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McStas_instr("test_instrument",
                              package_path=dummy_path,
                              input_path=input_path)

    os.chdir(current_work_dir)  # Return to previous workdir

    return instrument


def setup_instr_with_input_path_relative():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder. In addition the input_path
    is set to a folder in the test directory using a relative path.
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))

    current_work_dir = os.getcwd()
    os.chdir(THIS_DIR)  # Set work directory to test folder

    instrument = McStas_instr("test_instrument",
                              package_path="dummy_mcstas",
                              input_path="test_input_folder")

    os.chdir(current_work_dir)  # Return to previous workdir

    return instrument


def setup_populated_instr():
    """
    Sets up a neutron instrument with some features used and three components
    """
    instr = setup_instr_root_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    instr.add_component("first_component", "test_for_reading")
    instr.add_component("second_component", "test_for_reading")
    instr.add_component("third_component", "test_for_reading")

    return instr


def setup_populated_instr_with_dummy_path():
    """
    Sets up a neutron instrument with some features used and three components

    Here uses the dummy mcstas installation as path and sets required
    parameters so that a run is possible.
    """
    instr = setup_instr_with_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    comp1 = instr.add_component("first_component", "test_for_reading")
    comp1.gauss = 1.2
    comp1.test_string = "a_string"
    comp2 = instr.add_component("second_component", "test_for_reading")
    comp2.gauss = 1.4
    comp2.test_string = "b_string"
    comp3 = instr.add_component("third_component", "test_for_reading")
    comp3.gauss = 1.6
    comp3.test_string = "c_string"

    return instr


def setup_populated_x_ray_instr():
    """
    Sets up a X-ray instrument with some features used and three components
    """
    instr = setup_x_ray_instr_root_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    instr.add_component("first_component", "test_for_reading")
    instr.add_component("second_component", "test_for_reading")
    instr.add_component("third_component", "test_for_reading")

    return instr


def setup_populated_x_ray_instr_with_dummy_path():
    """
    Sets up a x-ray instrument with some features used and three components

    Here uses the dummy mcstas installation as path and sets required
    parameters so that a run is possible.
    """
    instr = setup_x_ray_instr_with_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    comp1 = instr.add_component("first_component", "test_for_reading")
    comp1.gauss = 1.2
    comp1.test_string = "a_string"
    comp2 = instr.add_component("second_component", "test_for_reading")
    comp2.gauss = 1.4
    comp2.test_string = "b_string"
    comp3 = instr.add_component("third_component", "test_for_reading")
    comp3.gauss = 1.6
    comp3.test_string = "c_string"

    return instr


def setup_populated_with_some_options_instr():
    """
    Sets up a neutron instrument with some features used and two components
    """
    instr = setup_instr_root_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    comp1 = instr.add_component("first_component", "test_for_reading")
    comp1.set_AT([0, 0, 1])
    comp1.set_GROUP("Starters")
    comp2 = instr.add_component("second_component", "test_for_reading")
    comp2.set_AT([0, 0, 2], RELATIVE="first_component")
    comp2.set_ROTATED([0, 30, 0])
    comp2.set_WHEN("1==1")
    comp2.yheight = 1.23
    instr.add_component("third_component", "test_for_reading")

    return instr


class TestDumpAndLoad(unittest.TestCase):
    def test_dump_simple(self):
        """
        Test a simple instrument can be dumped using with environment
        """
        my_instrument = setup_populated_instr_with_dummy_path()

        with WorkInTestDir() as handler:
            dump_name = "test_McStasScript_dump.dmp"

            if os.path.isfile(dump_name):
                os.remove(dump_name)

            # Ensure dump file does not exist
            self.assertFalse(os.path.isfile(dump_name))

            # Write dump file
            my_instrument.dump(dump_name)

            # Check it was written
            self.assertTrue(os.path.isfile(dump_name))

            # Delete dump file
            os.remove(dump_name)

    def test_load_simple(self):
        """
        Test a simple instrument can be loaded from dump
        """
        my_instrument = setup_populated_instr_with_dummy_path()
        my_instrument.add_parameter("check", comment="for testing")

        with WorkInTestDir() as handler:
            dump_name = "test_McStasScript_dump.dmp"
            my_instrument.dump(dump_name)

            loaded_instrument = McStas_instr.from_dump(dump_name)
            os.remove(dump_name)

            self.assertTrue(loaded_instrument, McStas_instr)
            self.assertTrue(loaded_instrument.name, my_instrument.name)

            self.assertEqual(len(loaded_instrument.component_list), 3)
            self.assertEqual(loaded_instrument.component_list[0].name, "first_component")
            self.assertEqual(loaded_instrument.component_list[2].name, "third_component")
            self.assertEqual(loaded_instrument.parameters["check"].name, "check")
            self.assertEqual(loaded_instrument.parameters["check"].comment, "for testing")
            self.assertEqual(loaded_instrument.parameters["has_default"].value, 37)




import os
import subprocess

def check_mcstas_major_version(mcstas_bin_path):
    """
    Checks installed McStas version
    """
    mcstas_command = os.path.join(mcstas_bin_path, "mcstas")
    output = subprocess.check_output([mcstas_command, "-v"])

    output = output.decode("utf-8")
    output = output.split(" (", 1)[0]
    output = output.split("version", 1)[1]
    major_version = output.split(".", 1)[0]

    return int(major_version)

def check_mcxtrace_major_version(mcstas_bin_path):
    """
    Checks installed McXtrace version
    """
    mcstas_command = os.path.join(mcstas_bin_path, "mcxtrace")
    output = subprocess.check_output([mcstas_command, "-v"])

    output = output.decode("utf-8")
    output = output.split(" (", 1)[0]
    output = output.split("version", 1)[1]
    major_version = output.split(".", 1)[0]

    return int(major_version)


from libpyvinyl.BaseFormat import BaseFormat
from mcstasscript.helper.managed_mcrun import load_results


class McStasFormat(BaseFormat):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def format_register(self):
        key = "sim"
        desciption = "sim format for McStasData"
        file_extension = ".sim"
        read_kwargs = [""]
        write_kwargs = [""]
        return self._create_format_register(
            key, desciption, file_extension, read_kwargs, write_kwargs
        )

    @staticmethod
    def direct_convert_formats():
        # Assume the format can be converted directly to the formats supported by these classes:
        # AFormat, BFormat
        # Redefine this `direct_convert_formats` for a concrete format class
        return []

    @classmethod
    def read(cls, filename: str) -> dict:
        """Read the data from the file with the `filename` to a dictionary. The dictionary will
        be used by its corresponding data class."""

        data = load_results(filename)
        data_dict = {"data": data}
        return data_dict

    @classmethod
    def write(cls, object, filename: str, key: str = None):
        """Don't have a way to write McStasData"""
        pass

    """
    #def write(cls, object: NumberData, filename: str, key: str = None):
        #Save the data with the `filename`.
        data_dict = object.get_data()
        arr = np.array([data_dict["number"]])
        np.savetxt(filename, arr, fmt="%.3f")
        if key is None:
            original_key = object.key
            key = original_key + "_to_TXTFormat"
        return object.from_file(filename, cls, key)
    """



from mcstasscript.instr_reader.util import SectionReader


class DefinitionReader(SectionReader):
    """
    Responsible for reading the defintion section of McStas instrument
    file. Contains instrument name and instrument parameters.
    """

    def __init__(self, Instr, write_file, product_filename,
                 get_next_line, return_line):

        super().__init__(Instr, write_file, product_filename,
                         get_next_line, return_line)

    def read_definition_line(self, line):
        """
        Reads line of instrument definition, returns bolean.  If it encounters
        the end of the definition section, it returns False, otherwise True.

        The contents of the definition section is written to the McStasScript
        Instr object.
        """

        continue_definition = True

        # Remove comments
        if "//" in line:
            line = line.split("//")[0]

        if "(" in line:
            # Start of instrument definition, get name
            self.instr_name = line.split("(")[0].strip().split(" ")[-1]
            self._start_py_file()
            # Remove the parameters from the paranthesis
            parameters = line.split("(")[1]
            if ")" in line:
                # Found end of definition
                continue_definition = False
                # these parameters are to be analyzed
                parameters = parameters.split(")")[0]

        elif ")" in line:
            # Found end of definition
            continue_definition = False
            # these parameters are to be analyzed
            parameters = line.split(")")[0]
        else:
            # Neither start or end on this line, analyze everything
            parameters = line

        # Separate into individual parameters
        parameters = parameters.split(",")
        if "\n" in parameters:
            parameters.remove("\n")

        for parameter in parameters:
            # Analyze individual parameter
            parameter = parameter.strip()

            if parameter == "":
                # If the parameter is empty, skip it.
                continue

            # Ready for keyword arguments
            kw_args = {}

            # Default to double type if nothing else is set
            parameter_type = "double"
            if " " and "=" in parameter:
                # Read parameter type
                type_and_name = parameter.split("=", 1)[0].strip()

                if " " in type_and_name:
                    parameter_type = type_and_name.split(" ", 1)[0].strip()
                    parameter = parameter.split(" ", 1)[1].strip()
            elif " " in parameter:
                # Read parameter type
                parameter_type = parameter.split(" ", 1)[0].strip()
                parameter = parameter.split(" ", 1)[1].strip()

            if "=" in parameter:
                # Read default value
                parameter_name = parameter.split("=")[0].strip()
                value = parameter.split("=")[1].strip()

                if parameter_type == "string":
                    if '"' in value:
                        pass
                        # Value has to be normal for object version
                        #value = value.replace('"', "\\\"")
                        #value = "\"" + value + "\""
                else:
                    if parameter_type == "int":
                        value = int(value)
                    else:
                        value = float(value)

                # Add defualt value to keyword arguments
                kw_args["value"] = value

            else:
                # No default value, just return the striped name
                parameter_name = parameter.strip()

            # Add this parameter to the object
            self.Instr.add_parameter(parameter_type, parameter_name, **kw_args)

            # Fix values for script version
            if parameter_type == "string" and "value" in kw_args:
                if isinstance(kw_args["value"], str):
                    kw_args["value"] = kw_args["value"].replace('"', '\\\"')
                    kw_args["value"] = "\"" + kw_args["value"] + "\""

            # Also write it to a file?
            write_string = []
            write_string.append(self.instr_name)
            write_string.append(".add_parameter(")
            write_string.append("\"" + parameter_type + "\"")
            write_string.append(", ")
            write_string.append("\"" + parameter_name + "\"")
            write_string.append(self._kw_to_string(kw_args))
            write_string.append(")\n")

            self._write_to_file(write_string)

        return continue_definition

    def _start_py_file(self):
        write_string = []

        # Write warning about robustness of this feature
        write_string.append("\"\"\"\n")
        write_string.append("This McStasScript file was generated from a\n")
        write_string.append("McStas instrument file. It is advised to check\n")
        write_string.append("the content to ensure it is as expected.\n")
        write_string.append("\"\"\"\n")

        # import McStasScript
        write_string.append("from mcstasscript.interface ")
        write_string.append("import ")
        write_string.append("instr, plotter, functions")
        write_string.append("\n\n")

        write_string.append(self.instr_name)
        write_string.append(" = instr.McStas_instr(")
        write_string.append("\"" + self.instr_name + "_generated\"")
        write_string.append(")\n")

        self._write_to_file(write_string)



import os
import sys
import unittest

from mcstasscript.interface import instr
from mcstasscript.instr_reader import control


def blockPrint():
    """
    Disables print capability
    """
    sys.stdout = open(os.devnull, 'w')


def enablePrint():
    """
    Restores print capability
    """
    sys.stdout = sys.__stdout__


def set_dummy_dir():
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    os.chdir(os.path.join(THIS_DIR, "dummy_instrument_folder"))


def setup_standard(Instr):
    set_dummy_dir()
    filename = "Union_demonstration_test.instr"
    InstrReader = control.InstrumentReader(filename)
    InstrReader.add_to_instr(Instr)

    return InstrReader


def setup_standard_auto_instr():
    set_dummy_dir()

    blockPrint()
    Instr = instr.McStas_instr("test_instrument")
    enablePrint()

    return setup_standard(Instr)


class TestInstrReader(unittest.TestCase):

    def test_read_instrument_name(self):
        """
        Check if the instrument name is read correctly
        """

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()

        filename = "Union_demonstration_test.instr"
        InstrReader = control.InstrumentReader(filename)
        InstrReader.add_to_instr(Instr)

        self.assertEqual(InstrReader.instr_name, "Union_demonstration")

    def test_read_input_parameter(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        parameter_list = list(Instr.parameters)

        self.assertEqual(parameter_list[0].name, "stick_displacement")
        # space in type inserted for easier writing by McStas_Instr class
        self.assertEqual(parameter_list[0].type, "double")
        self.assertEqual(parameter_list[0].value, 0)

        self.assertEqual(parameter_list[1].name, "test_int")
        # space in type inserted for easier writing by McStas_Instr class
        self.assertEqual(parameter_list[1].type, "int")
        self.assertEqual(parameter_list[1].value, 3)

        self.assertEqual(parameter_list[2].name, "test_str")
        # space in type inserted for easier writing by McStas_Instr class
        self.assertEqual(parameter_list[2].type, "string")
        self.assertEqual(parameter_list[2].value, "\"hurray\"")

    def test_read_declare_parameter(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        self.assertEqual(Instr.declare_list[0].name, "sample_1_index")
        self.assertEqual(Instr.declare_list[0].type, "int")
        self.assertEqual(Instr.declare_list[0].value, 27)

        self.assertEqual(Instr.declare_list[8].name, "array")
        self.assertEqual(Instr.declare_list[8].type, "double")
        self.assertEqual(Instr.declare_list[8].vector, 3)
        self.assertEqual(Instr.declare_list[8].value, [0.1, 0.2, 0.3])

        self.assertEqual(Instr.declare_list[9].name, "I_array")
        self.assertEqual(Instr.declare_list[9].type, "int")
        self.assertEqual(Instr.declare_list[9].vector, 4)

        self.assertEqual(Instr.declare_list[10].name, "T_array")
        self.assertEqual(Instr.declare_list[10].type, "int")
        self.assertEqual(Instr.declare_list[10].vector, 5)
        self.assertEqual(Instr.declare_list[10].value, [1, 2, 3, 4, 5])

        self.assertEqual(Instr.declare_list[11].name, "home")
        self.assertEqual(Instr.declare_list[11].type, "char")
        self.assertEqual(Instr.declare_list[11].vector, 20)
        self.assertEqual(Instr.declare_list[11].value,
                         "\\\"test_string\\\"")

    def test_read_initialize_line(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        initialize = ("// Start of initialize for generated test_instrument\n"
                      + "I_array[2] = 8;\n"
                      + "printf(\"Hello world\\n\");\n")

        self.assertEqual(Instr.initialize_section, initialize)

    # Check a few components are read correctly
    def test_read_component_1(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        components = Instr.component_list

        test_component = None

        for component in components:
            if component.name == "Al":
                test_component = component

        self.assertEqual(test_component.component_name, "Union_make_material")

        val = getattr(test_component, "my_absorption")
        self.assertEqual(val, "100*4*0.231/66.4")

        val = getattr(test_component, "process_string")
        self.assertEqual(val, "\"Al_incoherent,Al_powder\"")

        self.assertEqual(test_component.AT_data, ["0", "0", "0"])
        self.assertEqual(test_component.AT_relative, "ABSOLUTE")

        self.assertEqual(test_component.ROTATED_data, [0, 0, 0])
        self.assertEqual(test_component.ROTATED_relative, "ABSOLUTE")

    def test_read_component_2(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        components = Instr.component_list

        test_component = None

        for component in components:
            if component.name == "sample_holder3":
                test_component = component

        self.assertEqual(test_component.component_name, "Union_box")

        val = getattr(test_component, "xwidth")
        self.assertEqual(val, "0.0098")

        val = getattr(test_component, "priority")
        self.assertEqual(val, "52")

        self.assertEqual(test_component.AT_data,
                         ["0", "-0.03", "-0.03*0.35-0.004"])
        self.assertEqual(test_component.AT_relative,
                         "RELATIVE sample_rod_bottom")

        self.assertEqual(test_component.ROTATED_data, ["-25", "0", "0"])
        self.assertEqual(test_component.ROTATED_relative,
                         "RELATIVE sample_rod_bottom")

    def test_read_component_WHEN(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        components = Instr.component_list

        test_component = None

        for component in components:
            if component.name == "outer_cryostat_vacuum":
                test_component = component

        self.assertEqual(test_component.component_name, "Union_cylinder")

        val = getattr(test_component, "radius")
        self.assertEqual(val, "0.09")

        val = getattr(test_component, "priority")
        self.assertEqual(val, "11")

        self.assertEqual(test_component.WHEN, "WHEN (necessary == 1 )")

        self.assertEqual(test_component.AT_data, ["0", "0.01", "0"])
        self.assertEqual(test_component.AT_relative, "RELATIVE beam_center")

        self.assertEqual(test_component.ROTATED_data, ["0", "0", "0"])
        self.assertEqual(test_component.ROTATED_relative,
                         "RELATIVE beam_center")

    def test_read_component_EXTEND(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        components = Instr.component_list

        test_component = None

        for component in components:
            if component.name == "test_sample":
                test_component = component

        self.assertEqual(test_component.component_name, "Union_master")

        val = getattr(test_component, "history_limit")
        self.assertEqual(val, "1000000")

        lines = test_component.EXTEND.split("\n")
        line0 = ("if (scattered_flag[sample_1_index] > 0) scattered_1 = 1;"
                 + " else scattered_1 = 0;")
        line1 = ("if (scattered_flag[sample_2_index] > 0) scattered_2 = 1;"
                 + " else scattered_2 = 0;")
        line2 = ("if (scattered_flag[sample_3_index] > 0) scattered_3 = 1;"
                 + " else scattered_3 = 0;")
        line3 = ("if (scattered_flag[sample_4_index] > 0) scattered_4 = 1;"
                 + " else scattered_4 = 0;")

        self.assertEqual(lines[0], line0)
        self.assertEqual(lines[1], line1)
        self.assertEqual(lines[2], line2)
        self.assertEqual(lines[3], line3)

        self.assertEqual(test_component.AT_data, ["0", "0", "0"])
        self.assertEqual(test_component.AT_relative, "RELATIVE beam_center")

        self.assertEqual(test_component.ROTATED_data, ["0", "0", "0"])
        self.assertEqual(test_component.ROTATED_relative,
                         "RELATIVE beam_center")

    def test_read_component_GROUP(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        components = Instr.component_list

        test_component = None

        for component in components:
            if component.name == "armA":
                test_component = component

        self.assertEqual(test_component.component_name, "Arm")

        self.assertEqual(test_component.GROUP, "arms")

        self.assertEqual(test_component.AT_data, ["0", "0", "0"])
        self.assertEqual(test_component.AT_relative, "ABSOLUTE")

    def test_read_component_SPLIT(self):

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        components = Instr.component_list

        test_component = None

        for component in components:
            if component.name == "sample_4_container":
                test_component = component

        self.assertEqual(test_component.component_name, "Union_cylinder")

        self.assertEqual(test_component.AT_data, ["0", "0", "0"])
        self.assertEqual(test_component.AT_relative, "RELATIVE sample_4")

    def test_read_component_JUMP(self):
        """
        Check a JUMP and GROUP statement is read correctly
        """

        set_dummy_dir()

        blockPrint()
        Instr = instr.McStas_instr("test_instrument")
        enablePrint()
        setup_standard(Instr)

        components = Instr.component_list

        test_component = None

        for component in components:
            if component.name == "armB":
                test_component = component

        self.assertEqual(test_component.component_name, "Arm")

        self.assertEqual(test_component.GROUP, "arms")
        self.assertEqual(test_component.JUMP, "myself 2")

        self.assertEqual(test_component.AT_data, ["0", "0", "0"])
        self.assertEqual(test_component.AT_relative, "ABSOLUTE")

    def test_comma_split(self):
        """
        Test the Tracer_reader._split_func
        """

        InstrReader = setup_standard_auto_instr()

        test_string = "A,B,C,D(a,b),E"

        result = InstrReader.Trace_reader._split_func(test_string, ",")

        self.assertEqual(result[0], "A")
        self.assertEqual(result[1], "B")
        self.assertEqual(result[2], "C")
        self.assertEqual(result[3], "D(a,b)")
        self.assertEqual(result[4], "E")

    def test_comma_split_limited(self):
        """
        Test the Tracer_reader._split_func
        """

        InstrReader = setup_standard_auto_instr()

        test_string = "A,B,C,D(a,b),E"

        result = InstrReader.Trace_reader._split_func(test_string, ",",  2)

        self.assertEqual(result[0], "A")
        self.assertEqual(result[1], "B")
        self.assertEqual(result[2], "C,D(a,b),E")

    def test_parenthesis_split(self):
        """
        Test the Tracer_reader._split_func
        """

        InstrReader = setup_standard_auto_instr()

        test_string = "A)B)C)D(a,b))E"

        result = InstrReader.Trace_reader._split_func(test_string, ")")

        self.assertEqual(result[0], "A")
        self.assertEqual(result[1], "B")
        self.assertEqual(result[2], "C")
        self.assertEqual(result[3], "D(a,b)")
        self.assertEqual(result[4], "E")

    def test_comma_split_brack(self):
        """
        Test the Tracer_reader._split_func
        """

        InstrReader = setup_standard_auto_instr()

        test_string = "A,B{C,D(a,b)},E"

        result = InstrReader.Trace_reader._split_func_brack(test_string, ",")

        self.assertEqual(result[0], "A")
        self.assertEqual(result[1], "B{C,D(a,b)}")
        self.assertEqual(result[2], "E")


if __name__ == '__main__':
    unittest.main()



import copy

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from matplotlib.ticker import MaxNLocator
from matplotlib.colors import BoundaryNorm

from mcstasscript.data.data import McStasData
from mcstasscript.data.data import McStasDataEvent


def remove_eventdata(data_list, verbose=True):
    """
    Removes event data from a list, useful as these can't be plotted
    """
    reduced_data_list = []
    skipped_names = []
    for element in data_list:
        if not isinstance(element, McStasDataEvent):
            reduced_data_list.append(element)
        else:
            skipped_names.append(element.metadata.component_name)

    if verbose:
        for name in skipped_names:
            print(f"Skipped plotting {name} as it contains event data.")

    return reduced_data_list


def _fmt(x, pos):
    """
    Used for nice formatting of powers of 10 when plotting logarithmic
    """
    a, b = '{:.2e}'.format(x).split('e')
    b = int(b)
    if abs(float(a) - 1) < 0.01:
        return r'$10^{{{}}}$'.format(b)
    else:
        return r'${}\cdot 10^{{{}}}$'.format(a, b)


def _find_min_max_I(data):
    """
    Returns minimum and maximum intensity to plot given dataset

    Uses the plot options embedded in McStasData to determine the proper
    minimum and maximum intensity to display in a plot.

    Have to take cut_min and cut_max into account that can cut parts of
    the intensity away. When plotting logarithmic, orders_of_mags limits
    the orders of magnitude shown.

    Returns tuple of minimum and maximum, when no data is present the
    function returns 0, 0.
    """
    cut_max = data.plot_options.cut_max  # Default 1
    cut_min = data.plot_options.cut_min  # Default 0

    to_plot = data.Intensity

    min_value = to_plot.min()
    max_value = to_plot.max()

    if min_value == 0 and max_value == 0:
        return 0, 0

    if not data.plot_options.log:
        # Linear, simple case
        # Cut top and bottom of data as specified in cut variables
        min_value = min_value + (max_value - min_value) * cut_min
        max_value = max_value * cut_max

    else:
        # Logarithmic, minimum / maximum can not be zero
        max_data_value = to_plot.max()
        max_value = np.log10(max_data_value * cut_max)

        min_value = np.min(to_plot[np.nonzero(to_plot)])
        min_value = min_value + (max_data_value - min_value) * cut_min
        min_value = np.log10(min_value)

        # Take orders_of_mag into account (max / min in log10)
        if max_value - min_value > data.plot_options.orders_of_mag:
            min_value = max_value - data.plot_options.orders_of_mag

        # Convert back from log10
        min_value = 10.0 ** min_value
        max_value = 10.0 ** max_value

    return min_value, max_value


def _plot_fig_ax(data, fig, ax, **kwargs):
    """
    Plots the content of a single McStasData object

    Plotting is controlled through options associated with the
    McStasData objects.

    When plotting 2D objects, returns the pcolormesh object
    """

    if type(data.metadata.dimension) == int and data.metadata.dimension == 0:
        # Can't plot 0D data, show the info
        ax.text(0.5, 0.9, data.metadata.title, ha="center")
        ax.text(0.5, 0.7, "I: " + str(float(data.Intensity)), ha="center")
        ax.text(0.5, 0.5, "E: " + str(float(data.Error)), ha="center")
        ax.text(0.5, 0.3, "N: " + str(int(data.Ncount)), ha="center")
        ax.axis("off")

    elif type(data.metadata.dimension) == int and data.metadata.dimension != 0:

        x_axis_mult = data.plot_options.x_limit_multiplier

        x = data.xaxis * x_axis_mult
        y = data.Intensity
        y_err = data.Error

        ax.errorbar(x, y, yerr=y_err)

        ax.set_xlim(data.metadata.limits[0] * x_axis_mult,
                    data.metadata.limits[1] * x_axis_mult)

        # Add a title
        ax.set_title(data.metadata.title)

        # Add axis labels
        ax.set_xlabel(data.metadata.xlabel)
        ax.set_ylabel(data.metadata.ylabel)

        if data.plot_options.custom_xlim_left:
            ax.set_xlim(left=data.plot_options.left_lim)

        if data.plot_options.custom_xlim_right:
            ax.set_xlim(right=data.plot_options.right_lim)

        if data.plot_options.log:
            ax.set_yscale("log", nonpositive='clip')

            n_non_zero = np.count_nonzero(data.Intensity)
            if n_non_zero == 0:
                # Plot is empty, return
                return

            non_zero = np.nonzero(data.Intensity)
            min_value_log = np.log10(min(data.Intensity[non_zero]))
            max_value_log = np.log10(max(data.Intensity[non_zero]))

            orders_of_mag = data.plot_options.orders_of_mag
            if max_value_log - min_value_log > orders_of_mag:
                ax.set_ylim(top=1.1*10.0 ** max_value_log)
                ax.set_ylim(bottom=10.0 ** (max_value_log - orders_of_mag))

    elif len(data.metadata.dimension) == 2:

        min_value, max_value = _find_min_max_I(data)

        if "fixed_minimum_value" in kwargs:
            min_value = kwargs["fixed_minimum_value"]
        if "fixed_maximum_value" in kwargs:
            max_value = kwargs["fixed_maximum_value"]

        # Set the axis
        x_axis_mult = data.plot_options.x_limit_multiplier
        y_axis_mult = data.plot_options.y_limit_multiplier

        X = np.linspace(data.metadata.limits[0] * x_axis_mult,
                        data.metadata.limits[1] * x_axis_mult,
                        data.metadata.dimension[0] + 1)
        Y = np.linspace(data.metadata.limits[2] * y_axis_mult,
                        data.metadata.limits[3] * y_axis_mult,
                        data.metadata.dimension[1] + 1)

        # Create a meshgrid for both x and y
        x, y = np.meshgrid(X, Y)

        # Generate information on necessary colorrange
        levels = MaxNLocator(nbins=150).tick_values(min_value, max_value)

        # Select colormap
        cmap = copy.copy(plt.get_cmap(data.plot_options.colormap))
        if "no_data_to_black" in kwargs:
            if kwargs["no_data_to_black"]:
                cmap.set_bad((0, 0, 0))

        norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)

        # Empty data, return without cmap or norm
        if min_value == 0 and max_value == 0:
            levels = MaxNLocator(nbins=150).tick_values(0.001, 1.0)
            norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)
            im = ax.pcolormesh(x, y, data.Intensity, cmap=cmap, norm=norm)

        # Plot the data on the meshgrids
        elif data.plot_options.log:
            color_norm = matplotlib.colors.LogNorm(vmin=min_value,
                                                   vmax=max_value)
            im = ax.pcolormesh(x, y, data.Intensity,
                               cmap=cmap, norm=color_norm)
        else:
            im = ax.pcolormesh(x, y, data.Intensity, cmap=cmap, norm=norm)

        # Add the colorbar
        if data.plot_options.show_colorbar:
            cax = None
            if "colorbar_axes" in kwargs:
                cax = kwargs["colorbar_axes"]

            colorbar = fig.colorbar(im, ax=ax, cax=cax,
                                    format=matplotlib.ticker.FuncFormatter(_fmt))

            if data.metadata.zlabel is not None:
                colorbar.set_label(data.metadata.zlabel)

            if "colorbar_axes" in kwargs:
                cax.set_aspect(20)

        # Add a title
        ax.set_title(data.metadata.title)

        # Add axis labels
        ax.set_xlabel(data.metadata.xlabel)
        ax.set_ylabel(data.metadata.ylabel)

        if data.plot_options.custom_ylim_top:
            ax.set_ylim(top=data.plot_options.top_lim)

        if data.plot_options.custom_ylim_bottom:
            ax.set_ylim(bottom=data.plot_options.bottom_lim)

        if data.plot_options.custom_xlim_left:
            ax.set_xlim(left=data.plot_options.left_lim)

        if data.plot_options.custom_xlim_right:
            ax.set_xlim(right=data.plot_options.right_lim)

        return im
    else:
        print("Error, dimension not read correctly")


def _handle_kwargs(data_list, **kwargs):
    """
    Handle kwargs when list of McStasData objects given.

    Returns data_list

    data_list is turned into a list if it isn't already
    event data is removed as it can't be plotted directly

    Any kwargs can be given as a list, in that case apply them to given
    to the corresponding index.
    """

    if "fontsize" in kwargs:
        used_fontsize = kwargs["fontsize"]
    else:
        used_fontsize = 11
    plt.rcParams.update({'font.size': used_fontsize})

    if isinstance(data_list, McStasData):
        # Only a single element, put it in a list for easier syntax later
        data_list = [data_list]

    # Remove event data that can't be plotted in meaningful way
    data_list = remove_eventdata(data_list)

    known_plotting_kwargs = ["log", "orders_of_mag",
                             "top_lim", "bottom_lim",
                             "left_lim", "right_lim",
                             "cut_min", "cut_max",
                             "colormap", "show_colorbar",
                             "x_axis_multiplier",
                             "y_axis_multiplier"]

    for option in known_plotting_kwargs:
        if option in kwargs:
            given_option = kwargs[option]

            if isinstance(given_option, list):
                if len(data_list) < len(given_option):
                    raise ValueError("Keyword argument " + option + " is "
                                     + "given as a list, but this list has "
                                     + "more elements than there are "
                                     + "data sets to be plotted.")

                index = 0
                for per_list_option in given_option:
                    input_kwarg = {option: per_list_option}
                    data_list[index].set_plot_options(**input_kwarg)
                    index += 1

            else:
                for data in data_list:
                    input_kwarg = {option: given_option}
                    data.set_plot_options(**input_kwarg)

            # Remove option from kwargs
            del kwargs[option]

    return data_list


import os
from mcstasscript.instr_reader.control import InstrumentReader
from mcstasscript.interface.instr import McStas_instr


class McStas_file:
    """
    Reader of McStas files, can add to an existing McStasScript
    instrument instance or create a corresponding McStasScript python
    file.

    Methods
    -------

    add_to_instr(Instr)
        Add information from McStas file to McStasScript Instr instance

    write_python_file(filename)
        Write python file named filename that reproduce the McStas instr
    """

    def __init__(self, filename):
        """
        Initialization of McStas_file class, needs McStas instr filename

        Parameters
        ----------
            filename (str)
                Name of McStas instrument file to be read
        """

        # Check filename
        if not os.path.isfile(filename):
            raise ValueError("Given filename, \"" + filename
                             + "\" could not be found.")

        self.Reader = InstrumentReader(filename)

    def add_to_instr(self, Instr):
        """
        Adds information from the McStas file to McStasScript instr

        Parameters
        ----------
            Instr (McStasScript McStas_instr instance)
                McStas_instr instance to add instrument information to
        """

        # Check Instr
        if not isinstance(Instr, McStas_instr):
            raise TypeError("Given object is not of type McStas_instr!")

        self.Reader.add_to_instr(Instr)

    def write_python_file(self, filename, **kwargs):
        """
        Writes python file that reproduces McStas instrument file

        Parameters
        ----------
            filename (str)
                Filename of python file to be written
        """

        if "force" in kwargs:
            force = kwargs["force"]
        else:
            force = False

        # Check product_filename is available
        if os.path.isfile(filename):
            if force:
                os.remove(filename)
            else:
                raise ValueError("Filename \"" + filename
                                 + "\" already exists, you can overwrite with "
                                 + "force=True")

        self.Reader.generate_py_version(filename)
